/******************************************************************************
 *
 *	Multi-Channel USB audio (SINICHIRO) driver : sinich_drv.c
 *
 *	Copyright (c) 2006 by TED
 *
 *
 *	NOTES:
 *	 - the linked URBs would be preferred but not used so far because of
 *	   the instability of unlinking.
 */


/******************************************************************************
 *	include files
 */
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/fs.h>
#include <linux/string.h>
#include <linux/usb.h>
#include <linux/slab.h>
#include <linux/delay.h>

#include <asm/uaccess.h>

#include "sinich_cmn.h"
#include "sinich_prm.h"
#include "sinich_reg.h"
#include "sinich_lib.h"
#include "sinichdrv.h"


/******************************************************************************
 *	definitions
 */
// debug print level
//   0 : print no message
//   1 : print major debug messages
//   2 : print detail debug messages
#define SINICH_DRV_DEBUG 					0

// minor version number for /dev device file
#define SINICHIRO_DEVICE_MINOR				224

// default timeout (x10 msec) : 
#define SINICH_WAIT_DEFAULT_TIMEOUT			(HZ)

// URB buffer size
#define SINICHIRO_SYS_BLKIN01_BUF_SIZE		64

// limit number of URB submit
#define SINICH_URB_SUBMIT_LIMIT_NUMBER		10

//
#define 	MSG_SIZE	2*sizeof(uint)

//
#define		SMSG_SIZE	2*sizeof(uchar)


/******************************************************************************
 *	global variables
 */
static int vid[SNDRV_CARDS] = { [0 ... (SNDRV_CARDS-1)] = -1 }; /* Vendor ID for this card */
static int pid[SNDRV_CARDS] = { [0 ... (SNDRV_CARDS-1)] = -1 }; /* Product ID for this card */

unsigned long	audio_outpkt_tbl[7]={
					1536,	// 48k
					1024,	// 32k
					512,	// 16k
					256,	// 8k
					1536,	// 44.1k
					768,	// 22.05k
					384		// 11.025k
					};

unsigned long	audio_inpkt_tbl[7]={
					6144,	// 48k
					4096,	// 32k
					2048,	// 16k
					1024,	// 8k
					6144,	// 44k
					3072,	// 22k
					1536	// 11k
					};

static DECLARE_MUTEX(register_mutex);
static snd_SINICHIRO_t *usb_chip[SNDRV_CARDS];
static void SINICHIRO_ep01_complete_urb(struct urb *urb);
static void SINICHIRO_ep02h_complete_urb(struct urb *urb);
static void SINICHIRO_ep02_complete_urb(struct urb *urb);
static void SINICHIRO_ep04_complete_urb(struct urb *urb);
static int SINICHIRO_sys_stop(snd_SINICHIRO_t *chip, uint stream);

/******************************************************************************
 *	send control message
 *	send SINICHIRO vendor request
 *
 *	parameter :
 *		dev			: usb device instance pointer
 *		direction	: control transaction direction
 *		request		: request type
 *		fpga_addr	: FPGA address
 *		fpga_data	: FPGA data (write data)
 *		data		: read data pointer
 *		size		: read data size (bytes)
 *
 *	return value :
 *		success(0) or error(-1)
 */
static int SINICHIRO_vendor_request(snd_SINICHIRO_t *chip, uchar direction, uchar request,
						  ushort fpga_addr, ushort fpga_data,
						  void *data, ushort size)
{
	int transfer_size;
	uint pipe;

	struct usb_device *dev;	

#if (SINICH_DRV_DEBUG > 2)
	printk("SINICH-DRV : SINICHIRO_vendor_request\n");
#endif

	dev = chip->dev;	

	// calculate pipe
	pipe = (USB_DIR_OUT == direction) ? 
			usb_sndctrlpipe(dev,0) : usb_rcvctrlpipe(dev,0);

	// data copy to temporary buffer
	if (size > 0) memcpy((void *)&chip->vendorBuf, data, size);

	// call USB core driver function : usb_control_msg
	transfer_size = usb_control_msg(dev, pipe, request, USB_TYPE_VENDOR|direction,
					fpga_addr, fpga_data, &chip->vendorBuf, size, 10 * HZ);

	// free temporary buffer
	if (size > 0) memcpy(data, (void *)&chip->vendorBuf, size);

	// transfer size check
	if (transfer_size < 0)
	{
		printk("SINICH_DRV : SINICHIRO_vendor_request(%#x): ERROR : USB control transfer fail\n", request);
		return -1;
	}

	return 0;
}



/******************************************************************************
 *	audio or I/O control register write function
 *
 *	parameter :
 *		board_sel	: board select number (Audio board : 0-3, I/O board : 4-7)
 *		codec_sel	: codec chip select number : 0-3
 *		codec_addr	: codec control register address
 *		codec_data	: codec control register data
 *
 *	retrun value :
 *		success(0) or error(-1)
 */
#define SINICH_CODEC_CTL_TIMEOUT	10
static int SINICHIRO_codec_ctl_write(snd_SINICHIRO_t *chip, uchar board_sel, uchar codec_sel, uchar codec_addr, uchar codec_data)
{
	int err;
	ushort sys_addr = 0x0000;

#if (SINICH_DRV_DEBUG > 0)
	printk("SINICH-DRV : SINICHIRO_codec_ctl_write : chip ptr 0x%08X, board number 0x%02X, codec chip number 0x%02X\n",
		   (uint)chip, board_sel, codec_sel);
 	printk("SINICH-DRV : SINICHIRO_codec_ctl_write : codec addr 0x%02X, codec data ptr 0x%08X\n",
		   codec_addr, (uint)codec_data);
#endif


	// board number check
	if (0xF8 & board_sel) 
	{
		printk("SINICH-DRV : SINICHIRO_codec_ctl_write : ERROR : invalid board number 0x%02X\n", board_sel);
		return -1;
	}

	// codec chip number check
	if (0xFC & codec_sel)
	{
		printk("SINICH-DRV : SINICHIRO_codec_ctl_write : ERROR : invalid codec chip number 0x%02X\n", codec_sel);
		return -1;
	}

	// set codec high address
	//   Audio board #0 : 0x40
	//   Audio board #1 : 0x41
	//   Audio board #2 : 0x42
	//   Audio board #3 : 0x43
	//   I/O   board #0 : 0x44
	//   I/O   board #1 : 0x45
	//   I/O   board #2 : 0x46
	//   I/O   board #3 : 0x47

	// set codec address
	codec_addr |= (0x00F0 & (codec_sel << 4));

	// set write codec address
	sys_addr = 0x4000 | (0x0F00 & (board_sel << 8)) | 0x0001;
	err = SINICHIRO_vendor_request(chip, USB_DIR_OUT, VR_SET_FPGA_DATA, sys_addr, codec_addr, (void *)NULL, 0);
	if (err != 0)
	{
		printk("SINICH-DRV : SINICHIRO_codec_ctl_write : ERROR : codec address set fail\n");
		return -1;
	}

	// set codec data
	sys_addr = 0x4000 | (0x0F00 & (board_sel << 8)) | 0x0000;
	err = SINICHIRO_vendor_request(chip, USB_DIR_OUT, VR_SET_FPGA_DATA, sys_addr, codec_data, (void *)NULL, 0);
	if (err != 0)
	{
		printk("SINICH-DRV : SINICHIRO_codec_ctl_write : ERROR : codec data set fail\n");
		return -1;
	}

	// set write control codec data (0x01)
	sys_addr = 0x4000 | (0x0F00 & (board_sel << 8)) | 0x0002;
	err = SINICHIRO_vendor_request(chip, USB_DIR_OUT, VR_SET_FPGA_DATA, sys_addr, 0x01, (void *)NULL, 0);
	if (err != 0)
	{
		printk("SINICH-DRV : SINICHIRO_codec_ctl_write : ERROR : codec control data set fail\n");
		return -1;
	}

	// check interrupt factor bit and board ID
	{
		uchar tmp[1];
		uint i;

		sys_addr = 0x4800 | SINICH_REG_SYS_INT4_FACT;
		for (i = 0; ; i++)
		{
			// check timeout counter
			if (i >= SINICH_CODEC_CTL_TIMEOUT)
			{
				printk("SINICH-DRV : SINICHIRO_codec_ctl_write : ERROR : interrupt 4 factor write timeout\n");
				return -1;
			}

			// read INT4 factor register
			err = SINICHIRO_vendor_request(chip, USB_DIR_IN, VR_GET_FPGA_DATA, sys_addr, 0x00, (void *)tmp, 1);
			if (err != 0)
			{
				printk("SINICH-DRV : SINICHIRO_codec_ctl_write : ERROR : INT4 factor write fail\n");
				return -1;
			}

			// check WOK and ID
			if ((((0xF0 & *tmp) >> 4) == board_sel) && (0x01 & *tmp))
			{
				break;
			}
		}
	}

	// clear INT4 factor
	sys_addr = 0x4800 | SINICH_REG_SYS_INT4_FACT;
	err = SINICHIRO_vendor_request(chip, USB_DIR_OUT, VR_SET_FPGA_DATA, sys_addr, 0x00, (void *)NULL, 0);
	if (err != 0)
	{
		printk("SINICH-DRV : SINICHIRO_codec_ctl_write : ERROR : INT4 factor clear fail\n");
		return -1;
	}

	return 0;
}


/******************************************************************************
 *	audio or I/O control register read function
 *
 *	parameter :
 *		board_sel	: board select number (Audio board : 0-3, I/O board : 4-7)
 *		codec_sel	: codec chip select number : 0-3
 *		codec_addr	: codec control register address
 *		codec_data	: codec control register data pointer
 *
 *	retrun value : 
 *		success(0) or error(-1)
 */
static int SINICHIRO_codec_ctl_read(snd_SINICHIRO_t *chip, uchar board_sel, uchar codec_sel, uchar codec_addr, uchar *codec_data)
{
	int err;
	ushort sys_addr = 0x0000;

#if (SINICH_DRV_DEBUG > 0)
	printk("SINICH-DRV : SINICHIRO_codec_ctl_read : chip ptr 0x%08X, board number 0x%02X, codec chip number 0x%02X\n",
		   (uint)chip, board_sel, codec_sel);
 	printk("SINICH-DRV : SINICHIRO_codec_ctl_read : codec addr 0x%02X, codec data ptr 0x%08X\n",
		   codec_addr, (uint)codec_data);
#endif

	// board number check
	if (0xF8 & board_sel) 
	{
		printk("SINICH-DRV : SINICHIRO_codec_ctl_read : ERROR : invalid board number 0x%02X\n", board_sel);
		return -1;
	}

	// codec chip number check
	if (0xFC & codec_sel)
	{
		printk("SINICH-DRV : SINICHIRO_codec_ctl_read : ERROR : invalid codec chip number 0x%02X\n", codec_sel);
		return -1;
	}

	// set codec high address
	//   Audio board #0 : 0x40
	//   Audio board #1 : 0x41
	//   Audio board #2 : 0x42
	//   Audio board #3 : 0x43
	//   I/O   board #0 : 0x44
	//   I/O   board #1 : 0x45
	//   I/O   board #2 : 0x46
	//   I/O   board #3 : 0x47

	// set codec address
	codec_addr |= (0x00F0 & (codec_sel << 4));

	// set read codec address
	sys_addr = 0x4000 | (0x0F00 & (board_sel << 8)) | 0x0001;
	err = SINICHIRO_vendor_request(chip, USB_DIR_OUT, VR_SET_FPGA_DATA, sys_addr, codec_addr, (void *)NULL, 0);
	if (err != 0)
	{
		printk("SINICH-DRV : SINICHIRO_codec_ctl_read : ERROR : codec address set fail\n");
		return -1;
	}

	// set read control codec data (0x02)
	sys_addr = 0x4000 | (0x0F00 & (board_sel << 8)) | 0x0002;
	err = SINICHIRO_vendor_request(chip, USB_DIR_OUT, VR_SET_FPGA_DATA, sys_addr, 0x02, (void *)NULL, 0);
	if (err != 0)
	{
		printk("SINICH-DRV : SINICHIRO_codec_ctl_read : ERROR : codec control data set fail\n");
		return -1;
	}

	// check interrupt factor bit and board ID
	{
		uchar tmp[1];
		uint i;

		sys_addr = 0x4800 | SINICH_REG_SYS_INT4_FACT;
		for (i = 0; ; i++)
		{
			// check timeout counter
			if (i >= SINICH_CODEC_CTL_TIMEOUT)
			{
				printk("SINICH-DRV : SINICHIRO_codec_ctl_read : ERROR : interrupt 4 factor read timeout\n");
				return -1;
			}

			// read INT4 factor register
			err = SINICHIRO_vendor_request(chip, USB_DIR_IN, VR_GET_FPGA_DATA, sys_addr, 0x00, (void *)tmp, 1);
			if (err != 0)
			{
				printk("SINICH-DRV : SINICHIRO_codec_ctl_read : ERROR : INT4 factor read fail\n");
				return -1;
			}

			// check ROK and ID
			if ((((0xF0 & *tmp) >> 4) == board_sel) && (0x02 & *tmp))
			{
				break;
			}
		}
	}

	// clear INT4 factor
	sys_addr = 0x4800 | SINICH_REG_SYS_INT4_FACT;
	err = SINICHIRO_vendor_request(chip, USB_DIR_OUT, VR_SET_FPGA_DATA, sys_addr, 0x00, (void *)NULL, 0);
	if (err != 0)
	{
		printk("SINICH-DRV : SINICHIRO_codec_ctl_read : ERROR : INT4 factor clear fail\n");
		return -1;
	}

	// get codec data
	sys_addr = 0x4000 | (0x0F00 & (board_sel << 8)) | 0x0000;
	err = SINICHIRO_vendor_request(chip, USB_DIR_IN, VR_GET_FPGA_DATA, sys_addr, 0x00, (void *)codec_data, 1);
	if (err != 0)
	{
		printk("SINICH-DRV : SINICHIRO_codec_ctl_read : ERROR : codec data get fail\n");
		return -1;
	}

	return 0;
}


/******************************************************************************
 *	get FPGA data function
 *
 *	parameter :
 *		dev		: SINICHIRO USB device instance
 *		addr	: system board register address
 *		data	: system board register data pointer
 *
 *	return value :
 *		success(0) or error(-1)
 */
static int SINICHIRO_get_fpga_data(snd_SINICHIRO_t *chip, uchar addr, uchar *data)
{
	int err;
	uint sys_addr;

#if (SINICH_DRV_DEBUG > 0)
	printk("SINICH-DRV : SINICHIRO_get_fpga_data : dev ptr 0x%08X, addr 0x%02X\n", (uint)dev, addr);
#endif

	sys_addr = 0x4800 | (0xFF & addr);
	err = SINICHIRO_vendor_request(chip, USB_DIR_IN, VR_GET_FPGA_DATA, sys_addr, 0x00, (void *)data, 1);
	if (err != 0)
	{
		printk("SINICH-DRV : SINICHIRO_get_fpga_data : ERROR : FPGA data read fail\n");
		return -1;
	}

	return 0;
}

/******************************************************************************
 *	set FPGA data function
 *
 *	parameter :
 *		dev		: SINICHIRO USB device instance
 *		addr	: system board register address
 *		data	: system board register data pointer
 *
 *	return value :
 *		success(0) or error(-1)
 */
static int SINICHIRO_set_fpga_data(snd_SINICHIRO_t *chip, uchar addr, uchar data)
{
	int err;
	uint sys_addr;

#if (SINICH_DRV_DEBUG > 0)
	printk("SINICH-DRV : SINICHIRO_set_fpga_data : dev ptr 0x%08X, addr 0x%02X, data 0x%02X\n", (uint)dev, addr, data);
#endif

	sys_addr = 0x4800 | (0x00FF & addr);
	err = SINICHIRO_vendor_request(chip, USB_DIR_OUT, VR_SET_FPGA_DATA, sys_addr, data, (void *)NULL, 0);
	if (err != 0)
	{
		printk("SINICH-DRV : SINICHIRO_set_fpga_data : ERROR : FPGA data write fail\n");
		return -1;
	}

	return 0;
}


/******************************************************************************
 *	SINICHIRO stuck board information get function
 *
 *	parameter :
 *		chip		: SINICHIRO chip instance
 *		board_sel	: board select number : 0-7
 *		info		: codec control register "Act check" data pointer
 *
 *	return value :
 *		success(0) or error(-1)
 */
static int SINICHIRO_act_info(snd_SINICHIRO_t *chip, uchar board_sel, uchar *info)
{
	int err;

#if (SINICH_DRV_DEBUG > 0)
	printk("SINICH-DRV : SINICHIRO_act_info : chip ptr 0x%08X, board number 0x%02X, info ptr 0x%08X\n",
		   (uint)chip, board_sel, (uint)info);
#endif

	// call codec control read function
	// 0x00 : audio codec chip #0 or multi-I/O control #0
	err = SINICHIRO_codec_ctl_read(chip, board_sel, 0x00, SINICH_REG_CODEC_ACT_CHK, info);
	if (err != 0)
	{
		printk("SINICH-DRV : SINICHIRO_act_info : Act check data get fail\n");
		return -1;
	}

	return 0;
}


/******************************************************************************
 *	SINICHIRO system start function
 *	FPGA acitivate control
 *
 *	parameter :
 *		chip	: SINICHIRO chip instance pointer
 *		stream	: data stream
 *
 *	return value :
 *		success(0) or error(-1)
 */
static int SINICHIRO_sys_start(snd_SINICHIRO_t *chip, uint stream)
{
	int err = 0;

#if (SINICH_DRV_DEBUG > 0)
	printk("SINICH-DRV : SINICHIRO_sys_start : chip ptr 0x%08X\n", (uint)chip);
#endif

	switch (stream)
	{
		case SINICH_PCM_START:
		{
			//SINICHIRO_sys_stop(chip, SINICH_PCM_STOP);
			//msleep(2);
			// in
			err = SINICHIRO_vendor_request(chip, USB_DIR_OUT, VR_START_ITRANSFER, 0, 0, (void *)NULL, 0);
			if (err != 0)
			{
				return -1;
			}
			// out
			err = SINICHIRO_vendor_request(chip, USB_DIR_OUT, VR_START_OTRANSFER, 0, 0, (void *)NULL, 0);
			if (err != 0)
			{
				return -1;
			}
			err = SINICHIRO_vendor_request(chip, USB_DIR_OUT, VR_START_INTERRUPT, 0, 0, (void *)NULL, 0);
			if(err != 0){ // start interrupt transfer chain
				return -1;
			}
			{
				chip->rcvIndex=0;					// 
				chip->sndIndex=0;					// 
//				chip->rcvCount=0;		//061225 nakamura comment out
//				chip->wrCount=0;		//061225 nakamura comment out
				chip->rdIndex=0;					// 
				chip->wrIndex=0;					// 
				chip->tgl=0;
				chip->inpass=0;					// 
				chip->outpass=0;					// 
				chip->str_status=0;
				chip->snd=0;
				chip->chain_status=0x1;
				usb_fill_int_urb(chip->IntIn01urb, chip->dev, usb_rcvintpipe(chip->dev, 1),
						  &chip->pcm_status, 2,
						  (usb_complete_t)SINICHIRO_ep01_complete_urb, chip, 4);
				// submit URB
				err=usb_submit_urb(chip->IntIn01urb, GFP_ATOMIC);
			}
			if(err != 0) return -1;

			break;
		}
		case SINICH_PCM_CAPTURE :
		{
			// send vendor request
			err = SINICHIRO_vendor_request(chip, USB_DIR_OUT, VR_START_ITRANSFER, 0, 0, (void *)NULL, 0);

			// there is no error, set start flag
			if (err != 0)
			{
				return -1;
			}

			break;	
		}
		case SINICH_PCM_PLAYBACK :
		{
			// send vendor request
			err = SINICHIRO_vendor_request(chip, USB_DIR_OUT, VR_START_OTRANSFER, 0, 0, (void *)NULL, 0);

			// there is no error, set start flag
			if (err != 0)
			{
				return -1;
			}

			break;	
		}

		// Aug 17, 2006 :
		// -- del S--
#if 0
		case SINICH_IO_READ :
		{
			// send vendor request
			err = SINICHIRO_vendor_request(chip, USB_DIR_OUT, VR_IOSTART_ITRANSFER, 0, 0, (void *)NULL, 0);
			// there is no error, set start flag
			if (err == 0)
				return -1;
			}

			break;
		}
		case SINICH_IO_WRITE :
		{
			// send vendor request
			err = SINICHIRO_vendor_request(chip, USB_DIR_OUT, VR_IOSTART_OTRANSFER, 0, 0, (void *)NULL, 0);
			// there is no error, set start flag
			if (err == 0)
			{
				return -1;
			}

			break;
		}
#endif
		// -- del E--

		default :
		{
			printk("SINICH-DRV : SINICHIRO_sys_start : ERROR : stream invalid, 0x%2X\n", stream);
			return -1;
		}
	}

	return 0;
}


/******************************************************************************
 *	SINICHIRO system stop function
 *
 *	parameter :
 *		chip	: SINICHIRO chip instance pointer
 *		stream	: data stream
 *
 *	return value :
 *		success(0) or error(-1)
 */
static int SINICHIRO_sys_stop(snd_SINICHIRO_t *chip, uint stream)
{
	int err=0;

#if (SINICH_DRV_DEBUG > 0)
	printk("SINICH-DRV : SINICHIRO_sys_stop : chip ptr 0x%08X, stream %d\n",
		   (uint)chip, stream);
#endif


	switch (stream)
	{
		case SINICH_PCM_STOP:  // add for new ich
		{
			chip->chain_status = 0x00;
			msleep(2);
			chip->str_status = 0x0000;
			chip->rcvIndex=0;					// 
			chip->sndIndex=0;					// 
			chip->rcvCount=0;					// 
			chip->wrCount=0;					// 
			chip->rdIndex=0;					// 
			chip->wrIndex=0;					// 
			chip->tgl=0;
			err = SINICHIRO_vendor_request(chip, USB_DIR_OUT, VR_STOP_INTERRUPT, 0, 0, (void *)NULL, 0);
			if(err == 0)
				err = SINICHIRO_vendor_request(chip, USB_DIR_OUT, VR_STOP_ITRANSFER, 0, 0, (void *)NULL, 0);
			if(err == 0)
				err = SINICHIRO_vendor_request(chip, USB_DIR_OUT, VR_STOP_OTRANSFER, 0, 0, (void *)NULL, 0);
			break;
		}
		case SINICH_PCM_CAPTURE :
		{
			// send vendor request
			err = SINICHIRO_vendor_request(chip, USB_DIR_OUT, VR_STOP_ITRANSFER, 0, 0, (void *)NULL, 0);

			break;
		}
		case SINICH_PCM_PLAYBACK :
		{
			// send vendor request
			err = SINICHIRO_vendor_request(chip, USB_DIR_OUT, VR_STOP_OTRANSFER, 0, 0, (void *)NULL, 0);

			break;
		}

		// Aug 17, 2006 
		// -- del S--
#if 0
		case SINICH_IO_READ :
		{
			// send vendor request
			err = SINICHIRO_vendor_request(chip, USB_DIR_OUT, VR_IOSTOP_ITRANSFER, 0, 0, (void *)NULL, 0);

			break;
		}

		case SINICH_IO_WRITE :
		{
			// send vendor request
			err = SINICHIRO_vendor_request(chip, USB_DIR_OUT, VR_IOSTOP_OTRANSFER, 0, 0, (void *)NULL, 0);

			// set start flag
			chip->isDTstarted &= 0x07;

			break;
		}
#endif
		// -- del E--
		default :
		{
			printk("SINICH-DRV : SINICHIRO_sys_stop : ERROR : stream invalid, 0x%02X\n", stream);
			return -1;
		}
	}

	return err;
}


/******************************************************************************
 *	Endpoint 1 URB submit complete callback function
 *
 *	parameter :
 *		urb		: submitted URB structure pointer 
 *
 *	return value :
 *		nothing
 */
//static void SINICHIRO_ep01_complete_urb(struct urb *urb, struct pt_regs *regs)
static void SINICHIRO_ep01_complete_urb(struct urb *urb)
{
	snd_SINICHIRO_t *chip = urb->context;
	int status;

#if (SINICH_DRV_DEBUG > 2)
	printk("SINICH-DRV : SINICHIRO_ep01_complete_urb : urb pointer 0x%08X\n", (uint)urb);
#endif


	if(0x0==chip->chain_status) {
	/*	
		if(waitqueue_active(&chip->IntIn01EndNotify))
		{
			wake_up_interruptible(&chip->IntIn01EndNotify);
		}
	*/	
	} else {

		chip->str_status=chip->pcm_status;
		if( (chip->rcvCount < RNGBUF_NUM) && (chip->inpass == 0) ){// chk internal buffer status
			if(chip->str_status & 0x0F){// chk fpga fifo status
				// header
				usb_fill_bulk_urb(chip->BlkIn02urb[0], chip->dev, usb_rcvbulkpipe(chip->dev, 2),
								  chip->rcvBuf[chip->rcvIndex], 4,
								  (usb_complete_t)SINICHIRO_ep02h_complete_urb, chip);
				// submit URB
				status=usb_submit_urb(chip->BlkIn02urb[0], GFP_ATOMIC);
				if(status != 0) printk("submit error (In02urb[0])\n");	
/*
				// body
				usb_fill_bulk_urb(chip->BlkIn02urb[1], chip->dev, usb_rcvbulkpipe(chip->dev, 2),
								  chip->rcvBuf[chip->rcvIndex]+4, chip->inpkt_size,
								  (usb_complete_t)SINICHIRO_ep02_complete_urb, chip);
				chip->inpass=1;	
				// submit URB
				status=usb_submit_urb(chip->BlkIn02urb[1], GFP_ATOMIC);
				if(status != 0) printk("submit error (In02urb[1])\n");	
*/
			}//else{ printk("no valid fifo (ep2)\n"); }
		}//else{ printk("no sub %d %d\n",chip->rcvCount, chip->inpass);}
		  else {
			if( (chip->wrCount != 0) && (chip->outpass == 0) ){
				if(chip->str_status & 0x0f00){
					chip->outpass=1;
					usb_fill_bulk_urb(chip->BlkOut04urb, chip->dev, usb_sndbulkpipe(chip->dev, 4),
									  chip->sndBuf[chip->sndIndex], chip->outpkt_size,
									  (usb_complete_t)SINICHIRO_ep04_complete_urb, chip);
					// submit URB
					status=usb_submit_urb(chip->BlkOut04urb, GFP_ATOMIC);
					if(status != 0) printk("submit error (Out04urb\n");	
			
				}//else{ printk("no valid fito (ep4)\n"); }
			}
		}
		// nex interrupt transfer
		usb_fill_int_urb(chip->IntIn01urb, chip->dev, usb_rcvintpipe(chip->dev, 1),
			  &chip->pcm_status, 2,
			  (usb_complete_t)SINICHIRO_ep01_complete_urb, chip, 4);
		// submit URB
		status=usb_submit_urb(chip->IntIn01urb, GFP_ATOMIC);
		if(status != 0) printk("submit error (Int01urb\n");	

		/*
		if(chip->tgl){
			if(waitqueue_active(&chip->IntIn01TickNotify))
			{
				wake_up_interruptible(&chip->IntIn01TickNotify);
			}
			chip->tgl=0;
		} else { chip->tgl=1; }
		*/
	}

}


/******************************************************************************
 *	PCM sampling rate and channel number set function
 *
 *	parameter :
 *		dev		: USB device instance pointer
 *		data	: PCM sampling rate and channel number data
 *
 *	return value :
 *		err		: success(0) or error(-1)
 */
static int SINICHIRO_pcm_ctl(snd_SINICHIRO_t *chip, uchar data)
{
	int err;

#if (SINICH_DRV_DEBUG > 0)
	printk("SINICH-DRV : SINICHIRO_pcm_ctl : dev ptr 0x%08X, data0x%02X\n", (uint)dev, data);
#endif

	err = SINICHIRO_set_fpga_data(chip, SINICH_REG_SYS_AUDIO_SMPL_CTL, data);

	return err;
}


#if 0
/******************************************************************************
 *	I/O sampling rate and channel number set function
 *
 *	parameter :
 *		dev		: USB device instance pointer
 *		data	: I/O control data
 *
 *	return value :
 *		err		: success(0) or error(-1)
 */
static int SINICHIRO_io_ctl(struct usb_device *dev, uchar data)
{
	int err;

#if (SINICH_DRV_DEBUG > 0)
	printk("SINICH-DRV : SINICHIRO_io_ctl : dev ptr 0x%08X, data0x%02X\n", (uint)dev, data);
#endif

	err = SINICHIRO_set_fpga_data(dev, SINICH_REG_SYS_MCH_IO_SMPL_CTL, data);

	return err;
}
#endif

/******************************************************************************
 *	Endpoint 2 URB submit complete callback function
 *
 *	parameter :
 *		urb		: submitted URB structure pointer 
 *
 *	return value :
 *		nothing
 */
static void SINICHIRO_ep02h_complete_urb(struct urb *urb){

	int status;
	snd_SINICHIRO_t *chip = urb->context;
		chip->snd=1;
		if( (chip->wrCount != 0) && (chip->outpass == 0) ){
			
			if(chip->str_status & 0x0f00){
				chip->outpass=1;
				usb_fill_bulk_urb(chip->BlkOut04urb, chip->dev, usb_sndbulkpipe(chip->dev, 4),
								  chip->sndBuf[chip->sndIndex], chip->outpkt_size,
								  (usb_complete_t)SINICHIRO_ep04_complete_urb, chip);
				// submit URB
				status=usb_submit_urb(chip->BlkOut04urb, GFP_ATOMIC);
				if(status != 0) printk("submit error (Out04urb\n");	
			
			}//else{ printk("no valid fito (ep4)\n"); }
			
		}else{
		chip->snd=0;
	// body
	usb_fill_bulk_urb(chip->BlkIn02urb[1], chip->dev, usb_rcvbulkpipe(chip->dev, 2),
			  chip->rcvBuf[chip->rcvIndex]+4, chip->inpkt_size,
			  (usb_complete_t)SINICHIRO_ep02_complete_urb, chip);
	chip->inpass=1;	
	// submit URB
	status=usb_submit_urb(chip->BlkIn02urb[1], GFP_ATOMIC);
	if(status != 0) printk("submit error (In02urb[1])\n");	
	}
}
//static void SINICHIRO_ep02_complete_urb(struct urb *urb, struct pt_regs *regs)
static void SINICHIRO_ep02_complete_urb(struct urb *urb)
{
	snd_SINICHIRO_t *chip = urb->context;

#if (SINICH_DRV_DEBUG > 0)
	printk("SINICH-DRV : SINICHIRO_ep02_complete_urb : urb addr 0x%08X\n", (uint)urb);
#endif
	if(chip->BlkIn02urb[1] == urb){
		spin_lock(&chip->sts_lock);
		chip->rcvIndex++;
		chip->rcvIndex &= RNGBUF_MSK;
		chip->rcvCount++;
		chip->inpass=0;
		spin_unlock(&chip->sts_lock);
	}

}


/******************************************************************************
 *	Endpoint 4 URB submit complete callback function
 *
 *	parameter :
 *		urb		: submitted URB structure pointer 
 *
 *	return value :
 *		nothing
 */
//static void SINICHIRO_ep04_complete_urb(struct urb *urb, struct pt_regs *regs)
static void SINICHIRO_ep04_complete_urb(struct urb *urb)
{
	snd_SINICHIRO_t *chip = urb->context;
int status;
#if (SINICH_DRV_DEBUG > 0)
	printk("SINICH-DRV : SINICHIRO_ep04_complete_urb : urb addr 0x%08X\n", (uint)urb);
#endif

	spin_lock(&chip->sts_lock);
	chip->sndIndex++;
	chip->sndIndex &= RNGBUF_MSK;
	chip->wrCount--;
	chip->outpass=0;
	spin_unlock(&chip->sts_lock);


	if( chip->snd==1 ){
		chip->snd=0;
	// body
	usb_fill_bulk_urb(chip->BlkIn02urb[1], chip->dev, usb_rcvbulkpipe(chip->dev, 2),
			  chip->rcvBuf[chip->rcvIndex]+4, chip->inpkt_size,
			  (usb_complete_t)SINICHIRO_ep02_complete_urb, chip);
	chip->inpass=1;	
	// submit URB
	status=usb_submit_urb(chip->BlkIn02urb[1], GFP_ATOMIC);
	if(status != 0) printk("submit error (In02urb[1])\n");	
	}

}


/******************************************************************************
 *	audio or I/O control register read/write collectively function
 *	for one chip on the target board
 *
 *	parameter :
 *		chip		: SINICHIRO chip instance
 *		stream		: data stream
 *		board_sel	: board select number (Audio board : 0-3, I/O board : 4-7)
 *		codec_sel	: codec chip select number : 0-3
 *		tmpbuf		: temporary buffer pointer
 *		tmpbuf_size	: temporary buffer size
 *		tmpbuf_mask	: temporary buffer mask bit (read/write enable mask bit)
 *
 *	retrun value : 
 *		success(0) or error(-1)
 */
static int SINICHIRO_setget_codec(snd_SINICHIRO_t *chip, uchar stream, uchar board_sel,
							   uchar codec_sel, void *tmpbuf, uint tmpbuf_size, ushort tmpbuf_mask)
{
	int err;
	uchar *p, codec_addr;

#if (SINICH_DRV_DEBUG > 0)
	printk("SINICH-DRV : SINICHIRO_setget_codec : chip addr 0x%08X, board number 0x%02X, codec chip number 0x%02X, buffer size %d mask 0x%04x\n",
		   (uint)chip, board_sel, codec_sel, tmpbuf_size, tmpbuf_mask);
#endif

	// set pcm codec data
	for (codec_addr = 0x00, p = (uchar *)tmpbuf; codec_addr < tmpbuf_size-sizeof(tmpbuf_mask); codec_addr++, p++)
	{
		if (0x0001 & tmpbuf_mask)
		{
			if (0x01 & stream)
			{
				// call codec ctl write function
				err = SINICHIRO_codec_ctl_write(chip, board_sel, codec_sel, codec_addr, *p);
				if (err != 0)
				{
					printk("SINICH-DRV : SINICHIRO_setget_codec : ERROR : codec data set fail, codec addr 0x%02X, write data 0x%02X\n",
						   codec_addr, *p);
					return -1;
				}
			}
			else
			{
				// call codec ctl read function
				err = SINICHIRO_codec_ctl_read(chip, board_sel, codec_sel, codec_addr, p);
				if (err != 0)
				{
					printk("SINICH-DRV : SINICHIRO_setget_codec : ERROR : codec data set fail, codec addr 0x%02X, read data 0x%02X\n",
						   codec_addr, *p);
					return -1;
				}
			}
			// for debug
//			printk("SINICH-DRV : stream 0x%02X, board 0x%02X, codec 0x%02X, addr 0x%04X, data 0x%02X\n",
//				   stream, board_sel, codec_sel, codec_addr, *p);

		}
		tmpbuf_mask >>= 1;
	}

	return 0;
}


/******************************************************************************
 *	audio or I/O control register read/write collectively function
 *	for all chip on the target board
 *
 *	parameter :
 *		chip		: SINICHIRO chip instance
 *		stream		: data stream
 *		board_sel	: board select number (Audio board : 0-3, I/O board : 4-7)
 *		tmpbuf		: temporary buffer pointer
 *		tmpbuf_size	: temporary buffer size
 *
 *	retrun value : 
 *		success(0) or error(-1)
 */
static int SINICHIRO_setget_codec_all(snd_SINICHIRO_t *chip, uchar stream, uchar board_sel, void *tmpbuf, uint tmpbuf_size)
{
	int err;
	uchar codec_sel;
	ushort codec_member_mask;

#if (SINICH_DRV_DEBUG > 0)
	printk("SINICH-DRV : SINICHIRO_setget_codec_all : chip addr 0x%08X, stream 0x%02X, board number 0x%02X\n",
		   (uint)chip, stream, board_sel);
	printk("SINICH-DRV : SINICHIRO_setget_codec_all : buffer ptr 0x%08X, buffer size %d\n",
		   (uint)tmpbuf, tmpbuf_size);
#endif

	// check board select number
	if (0x02 & stream)
	{
		// get mask bit
		codec_member_mask = ((struct io_param_t *)tmpbuf)->act_flag;

		// I/O board check
		if (0xF8 & board_sel)
		{
			printk("SINICH-DRV : SINICHIRO_setget_codec_all : ERROR : invalid board select number 0x%02X\n", board_sel);
			return -1;
		}
	}
	else
	{
		// get mask bit
		codec_member_mask = ((struct codec_param_t *)tmpbuf)->act_flag;

		// PCM board check
		if (0xFC & board_sel)
		{
			printk("SINICH-DRV : SINICHIRO_setget_codec_all : ERROR : invalid board select number 0x%02X\n", board_sel);
			return -1;
		}
	}

	// audio or I/O control register read/write collectively for all chip on the target board 
	{
		uchar *codec_buffer;
		uint codec_buffer_size;

		// set temporary buffer start address
		codec_buffer = tmpbuf;

		// in write case, codec buffer size is equal to tmpbuf size
		// in read  case, codec buffer size is 1/4 size of tmpbuf size
		if (0x01 & stream)
		{
			codec_buffer_size = tmpbuf_size;
		}
		else
		{
			codec_buffer_size = tmpbuf_size >> 2;
		}

		// set pcm codec data
		for (codec_sel = 0; codec_sel < 4; codec_sel++)
		{
			// call codec ctl write function
			err = SINICHIRO_setget_codec(chip, stream, board_sel, codec_sel, codec_buffer, codec_buffer_size, codec_member_mask);
			if (err != 0)
			{
				printk("SINICH-DRV : SINICHIRO_setget_codec_all : ERROR : CODEC set/get fail\n");
				return -1;
			}

			if (0x01 & stream)
			{
				codec_buffer = tmpbuf;
			}
			else
			{
				codec_buffer += codec_buffer_size;
			}
		}
	}

	return 0;
}


/******************************************************************************
 *	SINICHIRO FPGA version number get function
 *
 *	parameter :
 *		chip		: SINICHIRO chip instance
 *		version		: version data pointer
 *
 *	return value :
 *		success(0) or error(-1)
 */
static int SINICHIRO_get_fpga_version(snd_SINICHIRO_t *chip, uint *version)
{
	int err;
	uchar tmpbuf[1];
	uint fpga_addr = 0;

#if (SINICH_DRV_DEBUG > 0)
	printk("SINICH-DRV : SINICHIRO_get_fpga_version: chip ptr 0x%08X\n", (uint)chip);
#endif

	// initialize temporary buffer
	tmpbuf[0] = '\0';

	// call vendor request
	fpga_addr = 0x4800 | SINICH_REG_SYS_FPGA_VERSION;
	err = SINICHIRO_vendor_request(chip, USB_DIR_IN, VR_GET_FPGA_DATA, fpga_addr, 0x00, (void *)tmpbuf, 1);
	if (err != 0)
	{
		printk("SINICH-DRV : SINICHIRO_get_fpga_version: ERROR : Get_fpga_version read failed\n");
		return -1;
	}

	// copy FPGA version data
	*version = (int)tmpbuf[0];

	return 0;
}


/******************************************************************************
 *	SINICHIRO device I/O control function
 *
 *	parameter :
 *		inode	: inode structure pointer
 *		file	: file structure pointer
 *		code	: I/O control command
 *		nbuf	: some data, data pointer to get or set
 *
 *	return value :
 *		err		: depends on I/O control command
 */
static int SINICHIRO_ioctl(struct inode *inode, struct file *file, uint code, unsigned long nbuf)
{
	int err = 0;
	ulong flag;
	snd_SINICHIRO_t *chip = (snd_SINICHIRO_t *)file->private_data;

#if (SINICH_DRV_DEBUG > 2)
	printk("SINICH-DRV : SINICHIRO_ioctl : node 0x%08X, code 0x%02X, nbuf 0x%08X\n",
		   (uint)inode, code, (uint)nbuf);
#endif


	if (NULL != chip)
	{
		switch (code)
		{
			// get FIFO status (EP01 bulk transaction)
			case SINICH_COM_SYS_STATUS :
			{
				msleep(2);
				//status = wait_event_interruptible_timeout(chip->IntIn01TickNotify, chip->str_status, HZ>>3);
				spin_lock_irqsave(&chip->sts_lock, flag);
				
				err = (((RNGBUF_NUM-chip->wrCount)<<8) | chip->rcvCount);
				
				spin_unlock_irqrestore(&chip->sts_lock, flag);

				break;
			}
			// PCM read (bulk read : EP02)
			case SINICH_COM_PCM_BLKIN :
			{
				uchar *ptmp;
				uint tmp[2];
				uint i, count, totalLength;

				// copy from user data, buffer pointer and buffer size, to kernel space
				err = copy_from_user(tmp, (void *)nbuf, MSG_SIZE);
				if(err != 0) { printk("copy_from_user error !\n"); return (-1);}

				// check the user memory for write
				if ( access_ok(VERIFY_WRITE, (void*)(char*)tmp[0], tmp[1])==0 ){
					//error
					printk("SINICH_DRV :buffer for write operation unvalid\n");
				}
				
				totalLength=0;
				if(chip->rcvCount){
					count=tmp[1]/chip->inpkt_size4;
					if(count > chip->rcvCount) count = chip->rcvCount;
					ptmp=(uchar *)tmp[0];	
					for( i=0 ; i<count ; i++ ){
						err=copy_to_user((void *)(ptmp), chip->rcvBuf[chip->rdIndex], chip->inpkt_size4);
//						if(err != chip->inpkt_size4){ printk("copy_to_user error !\n"); return (-1);}
						chip->rdIndex++;
						chip->rdIndex &= RNGBUF_MSK;
						totalLength+=chip->inpkt_size4;
						ptmp+=chip->inpkt_size4;
					}
					spin_lock_irqsave(&chip->sts_lock, flag);
					chip->rcvCount-=count;
					spin_unlock_irqrestore(&chip->sts_lock, flag);
				}
				err=totalLength;

				break;
			}
			// PCM write (bulk write : EP04)
			case SINICH_COM_PCM_BLKOUT :
			{
				uchar *ptmp;
				uint tmp[2];
				uint i,count, totalLength;

				// copy from user data, buffer pointer and buffer size, to kernel space
				err = copy_from_user(tmp, (void *)nbuf, MSG_SIZE);
				if(err != 0) { printk("copy_from_user error !\n"); return (-1);}
				
				// check the user memory for write
				if ( access_ok(VERIFY_READ, (void*)(char*)tmp[0], tmp[1])==0 ){
					//error
					printk("SINICH_DRV :buffer for write operation unvalid\n");
				}

				totalLength=0;
				if(chip->wrCount < RNGBUF_NUM){

					count=tmp[1]/chip->outpkt_size;
					if(count > (RNGBUF_NUM-chip->wrCount)) count = (RNGBUF_NUM-chip->wrCount);
					ptmp=(uchar *)tmp[0];
					for( i=0 ; i< count ; i++ ){
						err=copy_from_user(chip->sndBuf[chip->wrIndex], (void *)ptmp, chip->outpkt_size);
//						if(err != chip->outpkt_size){ printk("copy_from_user error !\n"); return (-1);}
						chip->wrIndex++;
						chip->wrIndex &= RNGBUF_MSK;
						totalLength+=chip->outpkt_size;
						ptmp+=chip->outpkt_size;
					}
					
					spin_lock_irqsave(&chip->sts_lock, flag);
					chip->wrCount+=count;
					spin_unlock_irqrestore(&chip->sts_lock, flag);
				}
				err=totalLength;

				break;
			}
			case SINICH_COM_PCM_READ :
			{
			
				break;
			}
			
			case SINICH_COM_PCM_WRITE :
			{
			
				break;
			}
			
			// get data from sinichiro system register
			case VR_GET_FPGA_DATA :
			{
				uint tmp[2];
				uchar tmpbuf[1];

				// copy from user data, buffer pointer and buffer size, to kernel space
				err = copy_from_user(tmp, (void *)nbuf, MSG_SIZE);
				if(err != 0) { printk("copy_from_user error !\n"); return (-1);}

				err = SINICHIRO_get_fpga_data(chip, tmp[0], tmpbuf);
				if (err != 0)
				{
					printk("SINICH-DRV : SINICHIRO_ioctl : ERROR : FPGA data read fail\n");
					return -1;
				}

				err = copy_to_user((void *)tmp[1], tmpbuf, sizeof(uchar));
				if(err != 0) { printk("copy_to_user error !\n"); return (-1);}

				break;
			}
			// set data to sinichiro system register
			case VR_SET_FPGA_DATA :
			{
				uchar tmp[2];

				// copy from user data, buffer pointer and buffer size, to kernel space
				err = copy_from_user(tmp, (void *)nbuf, SMSG_SIZE);
				if(err != 0) { printk("copy_from_user error !\n"); return (-1);}

				err = SINICHIRO_set_fpga_data(chip, tmp[0], tmp[1]);
				if (err != 0)
				{
					printk("SINICH-DRV : SINICHIRO_ioctl : ERROR : FPGA data read fail\n");
					return -1;
				}

				break;
			}

			case VR_ALLEP_START:
			{
				err = SINICHIRO_vendor_request(chip, USB_DIR_OUT, VR_ALLEP_START, 0, 0, (void *)NULL, 0);
				if (err != 0)
				{
					printk("SINICH-DRV : SINICHIRO_ioctl : ERROR : FPGA data write fail\n");
					return -1;
				}

				err = SINICHIRO_vendor_request(chip, USB_DIR_OUT, VR_START_INTERRUPT, 0, 0, (void *)NULL, 0);
				if (err != 0)
				{
					printk("SINICH-DRV : SINICHIRO_ioctl : ERROR : FPGA data write fail\n");
					return -1;
				}
				{// // start interrupt transfer chain

					/* 061225 nakamura add */
	                                chip->rcvIndex=0;                                       //
	                                chip->sndIndex=0;                                       //
	                                chip->rdIndex=0;                                        //
	                                chip->wrIndex=0;                                        //
	                                chip->tgl=0;
	                                chip->inpass=0;                                 //
	                                chip->outpass=0;                                        //
 	                                chip->snd=0;



					chip->str_status=0;
					chip->chain_status=0x1;
					usb_fill_int_urb(chip->IntIn01urb, chip->dev, usb_rcvintpipe(chip->dev, 1),
									&chip->pcm_status, 2,
									(usb_complete_t)SINICHIRO_ep01_complete_urb, chip, /*5*/4);
					// submit URB
					err=usb_submit_urb(chip->IntIn01urb, GFP_ATOMIC);
				}

				break;
			}
			case VR_ALLEP_STOP:
			{
				/* 061225 nakamura add */
	                        chip->chain_status = 0x00;
	                        msleep(2);
	                        chip->str_status = 0x0000;
	                        chip->rcvIndex=0;                                       //
	                        chip->sndIndex=0;                                       //
	                        chip->rcvCount=0;                                       //
	                        chip->wrCount=0;                                        //
	                        chip->rdIndex=0;                                        //
	                        chip->wrIndex=0;                                        //
	                        chip->tgl=0;

				err = SINICHIRO_vendor_request(chip, USB_DIR_OUT, VR_STOP_INTERRUPT, 0, 0, (void *)NULL, 0);
				if(err != 0)
				{
					printk("SINICH-DRV : SINICHIRO_ioctl : ERROR : FPGA data write fail\n");
					return -1;
				}
				err = SINICHIRO_vendor_request(chip, USB_DIR_OUT, VR_ALLEP_STOP, 0, 0, (void *)NULL, 0);
				if(err != 0)
				{
					printk("SINICH-DRV : SINICHIRO_ioctl : ERROR : FPGA data write fail\n");
					return -1;
				}

#if 0	//061225 nakamura comment out
				err = SINICHIRO_vendor_request(chip, USB_DIR_OUT, VR_ALLEP_STOP, 0, 0, (void *)NULL, 0);
				if (err != 0)
				{
					printk("SINICH-DRV : SINICHIRO_ioctl : ERROR : FPGA data write fail\n");
					return -1;
				}
			
				err = SINICHIRO_vendor_request(chip, USB_DIR_OUT, VR_STOP_INTERRUPT, 0, 0, (void *)NULL, 0);
				break;
#endif
			}			
			case SINICH_COM_SYS_ACT_INFO :
			{
				uint tmp[2];
				uchar *info=NULL;

				// copy from user data
				// tmp[0]	: board select number : 0-7
				// tmp[1]	: board data (Act check) pointer
				err = copy_from_user(tmp, (void *)nbuf, MSG_SIZE);
				if(err != 0) { printk("copy_from_user error !\n"); return (-1);}

				// memory allocate for "info"
				info = kmalloc(sizeof(uchar), GFP_KERNEL);
				if (!info)
				{
					printk("SINICH-DRV : SINICHIRO_ioctl : ERROR : no memory for \"info\"\n");
					return -ENOMEM;
				}

				// call function
				err = SINICHIRO_act_info(chip, tmp[0], info);

				// copy "Act check" data to user space 
				err = copy_to_user((void *)tmp[1], (void *)info, sizeof(uchar));
				if(info != NULL) kfree(info);
				if(err != 0) { printk("copy_to_user error !\n"); return (-1);}

				break;
			}
			case SINICH_COM_SYS_START :
			{
				// send vendor request : VR_START
				err = SINICHIRO_sys_start(chip, (uint)nbuf);
				if (err != 0)
				{
					printk("SINICH-DRV : SINICHIRO_ioctl : ERROR : system start fail\n");
					return -1;
				}

				break;
			}
			case SINICH_COM_SYS_STOP :
			{
				// send vendor request : VR_STOP
				err = SINICHIRO_sys_stop(chip, (uint)nbuf);
				if (err != 0)
				{
					printk("SINICH-DRV : SINICHIRO_ioctl : ERROR : system stop fail\n");
					return -1;
				}

				break;
			}
			case SINICH_COM_PCM_CTL :
			{
				// call function
				err = SINICHIRO_pcm_ctl(chip, (uchar)nbuf);
				if (err != 0)
				{
					printk("SINICH-DRV : SINICHIRO_ioctl : ERROR : PCM sampling rate, channel set fail\n");
					return -1;
				}else{
					chip->inpkt_size  = audio_inpkt_tbl[(uchar)(nbuf&0x0F)];
					chip->inpkt_size4 = chip->inpkt_size+4;
					chip->outpkt_size = audio_outpkt_tbl[(uchar)(nbuf&0x0F)];
					chip->rcvCount=0;	//061225 nakamura add
					chip->wrCount=0;	//061225 nakamura add
				}

				break;
			}
			case SINICH_COM_PCM_SET_CODEC :
			{
				uint tmp[3];
				uint tmpbuf_size;
				uchar *tmpbuf=NULL;

				// copy from user data to kernel memory space :  buffer pointer and buffer size
				err = copy_from_user(tmp, (void *)nbuf, 3*sizeof(uint));
				if(err != 0) { printk("copy_from_user error !\n"); return (-1);}

				// set stream SET bit
				// bit 1 : PCM stream(0) or I/O stream(1)
				// bit 0 : read(0) or write(1)
				if (SINICH_COM_PCM_SET_CODEC == code)
				{
					tmp[0] = 0x01;
				}
				else
				{
					tmp[0] = 0x03;
				}

				// calculate temporary buffer size
				if(0x02 & (uchar)tmp[0])
				{
					tmpbuf_size = sizeof(struct io_param_t);
				}
				else
				{
					tmpbuf_size = sizeof(struct codec_param_t);
				}

				// temporary buffer memory allocate
				tmpbuf = kmalloc(tmpbuf_size, GFP_KERNEL);
				if (!tmpbuf)
				{
					printk("SINICH-DRV : SINICHIRO_ioctl : ERROR : no memory for audio, I/O control write\n");
					return -1;
				}

				// copy from user data to kernel memory space : audio codec data
				err = copy_from_user(tmpbuf, (void *)tmp[2], tmpbuf_size);
				if(err != 0) { printk("copy_from_user error !\n");
				} else {
				// call specific function
					err = SINICHIRO_setget_codec_all(chip, tmp[0], tmp[1], (void *)tmpbuf, tmpbuf_size);
					if (err != 0)
					{
						printk("SINICH-DRV : SINICHIRO_ioctl : ERROR : audio, I/O control write fail\n");
					}
				}

				// free memory
				if(tmpbuf != NULL) kfree(tmpbuf);
				if(err != 0) return (-1);
				break;
			}
			case SINICH_COM_PCM_GET_CODEC :
			{
				uint tmp[3];
				uint tmpbuf_size;
				uchar *tmpbuf;


				// copy from user data to kernel memory space :  buffer pointer and buffer size
				err = copy_from_user(tmp, (void *)nbuf, 3*sizeof(uint));
				if(err != 0) { printk("copy_from_user error !\n"); return (-1);}

				// set stream SET bit
				// bit 1 : PCM stream(0) or I/O stream(1)
				// bit 0 : read(0) or write(1)
				if (SINICH_COM_PCM_GET_CODEC == code)
				{
					tmp[0] = 0x00;
				}
				else
				{
					tmp[0] = 0x02;
				}

				// calculate temporary buffer size
				if(0x02 & (uchar)tmp[0])
				{
					tmpbuf_size = 4*sizeof(struct io_param_t);
				}
				else
				{
					tmpbuf_size = 4*sizeof(struct codec_param_t);
				}

				// temporary buffer memory allocate
				tmpbuf = kmalloc(tmpbuf_size, GFP_KERNEL);
				if (!tmpbuf)
				{
					printk("SINICH-DRV : SINICHIRO_ioctl : ERROR : no memory for audio, I/O control read\n");
					return -1;
				}

				// copy from user data to kernel memory space : audio codec data
				err = copy_from_user(tmpbuf, (void *)tmp[2], tmpbuf_size);
				if(err != 0) { printk("copy_from_user error !\n"); return (-1);}

#if (SINICH_DRV_DEBUG > 1)
				printk("SINICH-DRV : SINICHIRO_ioctl : XXX_GET_CODEC : temporary buffer size 0x%02X\n", tmpbuf_size);
				printk("SINICH-DRV : SINICHIRO_ioctl : XXX_GET_CODEC : tmp[0] 0x%02X, tmp[1] 0x%02X, tmp[2] 0x%04x\n", tmp[0], tmp[1], tmp[2]);
#endif

				// call specific function
				err = SINICHIRO_setget_codec_all(chip, tmp[0], tmp[1], (void *)tmpbuf, tmpbuf_size);
				if (err != 0)
				{
					printk("SINICH-DRV : SINICHIRO_ioctl : ERROR : audio, I/O control read fail\n");
					return -1;
				}

				// copy to user space : audio codec data
				err = copy_to_user((void *)tmp[2], tmpbuf, tmpbuf_size);
				if(err != 0) { printk("copy_to_user error !\n"); return (-1);}

				// free memory
				if(tmpbuf != NULL) kfree(tmpbuf);

				break;

			}
			// CODEC(audio, I/O) control data read
			case SINICH_COM_SYS_GET_CODEC_CTL : 
			{
				uint tmp[4];
				uchar buf[1];

				// copy from user data, buffer pointer and buffer size, to kernel space
				// tmp[0]	: board select number
				// tmp[1]	: codec select number
				// tmp[2]	: codec control register address
				// tmp[3]	: codec control register data pointer
				err = copy_from_user(tmp, (void *)nbuf, 4*sizeof(uint));
				if(err != 0) { printk("copy_from_user error !\n"); return (-1);}

				// CODEC control data read
				err = SINICHIRO_codec_ctl_read(chip, tmp[0], tmp[1], tmp[2], buf);
				if (err != 0)
				{
					printk("SINICH-DRV : SINICHIRO_ioctl : ERROR : codec control data get fail\n");
					return -1;
				}

				// copy CODEC control data to user space
				err = copy_to_user((void *)tmp[3], buf, sizeof(uchar));
				if(err != 0) { printk("copy_to_user error !\n"); return (-1);}

				break;
			}
			// CODEC (audio, I/O) control data write
			case SINICH_COM_SYS_SET_CODEC_CTL :
			{
				uchar tmp[4];

				// copy from user data, buffer pointer and buffer size, to kernel space
				// tmp[0]	: board select number
				// tmp[1]	: codec select number
				// tmp[2]	: codec control register address
				// tmp[3]	: codec control register data
				err = copy_from_user(tmp, (void *)nbuf, 4*sizeof(uchar));
				if(err != 0) { printk("copy_from_user error !\n"); return (-1);}

				// CODEC control data write
				err = SINICHIRO_codec_ctl_write(chip, tmp[0], tmp[1], tmp[2], tmp[3]);
				if (err != 0)
				{
					printk("SINICH-DRV : SINICHIRO_ioctl : ERROR : codec control data set fail\n");
					return -1;
				}

				break;
			}
			// this device driver version get
			case SINICH_COM_SYS_GET_FPGA_VERSION :
			{
				uint fpga_version[1];

				err = SINICHIRO_get_fpga_version(chip, fpga_version);
				if (err != 0)
				{
					printk("SINICH-DRV : SINICHIRO_ioctl : ERROR : FPGA version read fail\n");
					return -1;
				}

				err = fpga_version[0];
				printk("fpga version 0x%04X\n", err);
				break;
			}
			// this device driver version get
			case SINICH_COM_SYS_GET_DRIVER_VERSION :
			{
				// set "driver version" to return value directly
				err = SINICH_DRIVER_VERSION;

				break;
			}
			// get USB devcie descriptor
			case SINICH_COM_SYS_GET_USB_DEVICE_DESCRIPTOR :
			{
				uint tmp[1];

				// copy from user data, buffer pointer and buffer size, to kernel space
				// tmp[0]	: USB device descriptor structure pointer
				err = copy_from_user(tmp, (void *)nbuf, sizeof(uint));
				if(err != 0) { printk("copy_from_user error !\n"); return (-1);}

				// copy CODEC control data to user space
				err = copy_to_user((void *)tmp[0], &(chip->dev->descriptor), sizeof(USB_DEVICE_DESCRIPTOR));
				if(err != 0) { printk("copy_to_user error !\n"); return (-1);}

				break;
			}
			// invalid command
			default :
			{
				printk("SINICH-DRV : SINICHIRO_ioctl : ERROR : no such command\n");
				return -1;
			}
		}
	}
	else
	{
		printk("SINICH-DRV : SINICHIRO_ioctl : ERROR : no chip instance\n");
		return -1;
	}

	return err;
}


/******************************************************************************
 *	find SINICHIRO instance function
 *
 *	parameter :
 *		nothing
 *
 *	return value :
 *		SINICHIRO instance pointer
 */
static snd_SINICHIRO_t *find_SINICHIRO_instance(void)
{
	uint i;

#if (SINICH_DRV_DEBUG > 1)
	printk("SINICH-DRV : find_SINICHIRO_instance\n");
#endif

	for (i = 0; i < SNDRV_CARDS; i++)
	{
		if (NULL != usb_chip[i]) 
		{
			return usb_chip[i];
		}
	}

	return NULL;
}


/******************************************************************************
 *	SINICHIRO device free function
 *	free alocated memory
 *
 *	parameter :
 *		chip	: SINICHIRO chip instance pointer
 *
 *	return value :
 *		nothing
 */
static void SINICHIRO_free_instance(snd_SINICHIRO_t *chip)
{
	uint i;


	if(chip != NULL){
		// de-allocate URB
		if (chip->IntIn01urb != NULL)    { usb_free_urb(chip->IntIn01urb); chip->IntIn01urb = NULL;}
		if (chip->BlkIn02urb[0] != NULL) { usb_free_urb(chip->BlkIn02urb[0]); chip->BlkIn02urb[0] = NULL;}
		if (chip->BlkIn02urb[1] != NULL) { usb_free_urb(chip->BlkIn02urb[1]); chip->BlkIn02urb[1] = NULL;}
		if (chip->BlkOut04urb != NULL)   { usb_free_urb(chip->BlkOut04urb); chip->BlkOut04urb = NULL;}
		for( i=0; i<RNGBUF_NUM ; i++ ){
			if (chip->rcvBuf[i] != NULL) { kfree(chip->rcvBuf[i]);}
			if (chip->sndBuf[i] != NULL) { kfree(chip->sndBuf[i]);}
		}
	}
}


/******************************************************************************
 *	SINICHIRO device open function
 *	initialize SINICHIRO instance
 *
 *	parameter :
 *		inode	: inode structure pointer
 *		file	: file structure pointer
 *
 *	return value :
 */
static int SINICHIRO_open(struct inode *inode, struct file *file)
{
	snd_SINICHIRO_t *chip = find_SINICHIRO_instance();

#if (SINICH_DRV_DEBUG > 0)
	printk("SINICH-DRV : SINICHIRO_open : i-node ptr 0x%08X, file ptr 0x%08X\n", (uint)inode, (uint)file);
#endif

	// check device instance
	if (NULL == chip)
	{
		printk("SINICH-DRV : No SINICHIRO device\n");
		return -ENODEV;
	}

	// open counter increment
	chip->isopen++;
	// set chip instance to file private data structure
	file->private_data = chip;
	chip->chain_status=0x0;
	chip->inpass=0;
	chip->outpass=0;

	return 0;

}


/******************************************************************************
 *	SINICHIRO device close function
 *	free SINICHIRO instance
 *
 *	parameter :
 *		inode	: inode structure pointer
 *		file	: file structure pointer
 *
 *	return value :
 *		always success(0)
 */
static int SINICHIRO_close(struct inode *inode, struct file *file)
{
	snd_SINICHIRO_t *chip = (snd_SINICHIRO_t *)file->private_data;

#if (SINICH_DRV_DEBUG > 0)
	printk("SINICH-DRV : SINICHIRO_close : inode addr 0x%08X, file addr 0x%08X\n",
		   (uint)inode, (uint)file);
#endif

	SINICHIRO_sys_stop(chip, SINICH_PCM_STOP);

	// open counter decrement
	chip->isopen--;

	return 0;
}


/******************************************************************************
 *	create SINICHIRO chip instance function
 *
 *	parameter :
 *		dev			: usb device instance pointer
 *		idx			: index number
 *		rchip		: SINICHIRO chip instance pointer pointer
 *
 *	return value :
 *		success(0) or error(-1)
 */
static int snd_SINICHIRO_create(struct usb_device *dev, int idx, snd_SINICHIRO_t **rchip)
{
	int i;
	snd_SINICHIRO_t *chip;

#if (SINICH_DRV_DEBUG > 0)
	printk("SINICH-DRV : snd_SINICHIRO_create : device ptr 0x%08X, ID %d, chip ptr 0x%08X\n",
		   (uint)dev, idx, (uint)rchip);
#endif

	// initialize chip instance
	*rchip = NULL;

	// SINICHIRO supports only High Speed
	if (snd_usb_get_speed(dev) != USB_SPEED_HIGH) 
	{
		printk("SINICH-DRV : snd_SINICHIRO_create : SINICHIRO supports only High Speed %d\n", snd_usb_get_speed(dev));
		return -1;
	}

	// kernel memory allocate for SINICHIRO and clear the region
	chip = kmalloc(sizeof(snd_SINICHIRO_t), GFP_KERNEL);
	if (! chip)
	{
		printk("SINICH-DRV : snd_SINICHIRO_create : cannot kernel memory allocate for chip\n");
		return -1;
	}
	memset(chip, (uchar)0, sizeof(snd_SINICHIRO_t));

	chip->index = idx;
	chip->dev = dev;

	init_waitqueue_head(&chip->IntIn01TickNotify);
	init_waitqueue_head(&chip->IntIn01EndNotify);

	spin_lock_init(&chip->sts_lock);

	// allocate URB end point 01
	if (!(chip->IntIn01urb  = usb_alloc_urb(0, GFP_KERNEL)))
	{
		printk("SINICH-DRV : SINICHIRO_create : ERROR : EP01 usb_alloc_urb fail\n");
		goto __SINICH_RESOURCE_ERROR;
	}

	// allocate URB : end point 02
	if (!(chip->BlkIn02urb[0] = usb_alloc_urb(0, GFP_KERNEL)))
	{
		printk("SINICH-DRV : SINICHIRO_create : ERROR : EP02 usb_alloc_urb fail\n");
		goto __SINICH_RESOURCE_ERROR;
	}

	// allocate URB : end point 02
	if (!(chip->BlkIn02urb[1] = usb_alloc_urb(0, GFP_KERNEL)))
	{
		printk("SINICH-DRV : SINICHIRO_create : ERROR : EP02 usb_alloc_urb fail\n");
		goto __SINICH_RESOURCE_ERROR;
	}

	if (!(chip->BlkOut04urb  = usb_alloc_urb(0, GFP_KERNEL)))
	{
		printk("SINICH-DRV : SINICHIRO_create : ERROR : EP04 usb_alloc_urb fail\n");
		goto __SINICH_RESOURCE_ERROR;
	}

	memset(chip->rcvBuf,0,sizeof(chip->rcvBuf));
	memset(chip->sndBuf,0,sizeof(chip->sndBuf));
	for( i=0; i<RNGBUF_NUM; i++ ){
		chip->rcvBuf[i]=kmalloc(6148, GFP_KERNEL);
		if(chip->rcvBuf[i] == NULL){
			printk("SINICH-DRV : SINICHIRO_create : ERROR : kmalloc for EP02 read buffer fail\n");
			goto __SINICH_RESOURCE_ERROR;
		}
		chip->sndBuf[i]=kmalloc(1536, GFP_KERNEL);
		if(chip->sndBuf[i] == NULL){
			printk("SINICH-DRV : SINICHIRO_create : ERROR : kmalloc for EP04 write buffer fail\n");
			goto __SINICH_RESOURCE_ERROR;
		}
	}
	*rchip = chip;

	return (0);

__SINICH_RESOURCE_ERROR:

	SINICHIRO_free_instance(chip);
	
	return (-1);
}


/******************************************************************************
 *	USB control transfer : reset configuration
 *
 *	parameter :
 *		dev			: usb device instance pointer
 *
 *	return value :
 *		usb_set_configuration
 */
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2, 5, 70)	/* FIXME: find correct number */
static int usb_reset_configuration(struct usb_device *dev)
{
//	struct usb_host_config *config = dev->actconfig;
	struct usb_config_descriptor *config = dev->actconfig;

	/* FIXME: replace with something more useful like in 2.6 code? */
	return usb_set_configuration(dev, get_cfg_desc(config)->bConfigurationValue);
}
#endif



/******************************************************************************
 *	USB file operation structure
 */
static struct file_operations SINICHIRO_fops = 
{
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2, 5, 70)	/* FIXME: find right number */
	.owner			= THIS_MODULE,
#endif
	.ioctl			= SINICHIRO_ioctl,
	.open			= SINICHIRO_open,
	.release		= SINICHIRO_close,
};

/******************************************************************************
 *	USB class driver structure
 */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 5, 70)	/* FIXME: find right number */
static struct usb_class_driver SINICHIRO_class =
{
	.name			= "usb/sinichusb0",
	.fops			= &SINICHIRO_fops,
//	.mode			= S_IFCHR | S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH,
    .minor_base		= SINICHIRO_DEVICE_MINOR,
};
#endif



/******************************************************************************
 *	SINICHIRO probe function
 *
 *	we check the usb device pointer and creates the card instance
 *	only at the first time.  the successive calls of this function will
 *	append the pcm interface to the corresponding card.
 *
 *	parameter :
 *		dev			: usb device instance pointer
 *		intf		: usb interface structure pointer
 *		usb_id		: usb ID number
 *
 *	return value :
 *		SINICHIRO chip instance pointer
 */
static void *snd_SINICHIRO_probe(struct usb_device *dev, struct usb_interface *intf, const struct usb_device_id *usb_id)
{
	int i, err;
	snd_SINICHIRO_t *chip;


	printk("SINICH_DRV : enter snd_SINICHIRO_probe %d\n", HZ);
	// check whether it's already registered
	chip = NULL;
	down(&register_mutex);
	for (i = 0; i < SNDRV_CARDS; i++)
	{
		if (usb_chip[i] && usb_chip[i]->dev == dev)
		{
			if (usb_chip[i]->shutdown)
			{
				printk("SINICH-DRV : USB device is in the shutdown state, cannot create a card instance\n");
				goto __error;
			}
			chip = usb_chip[i];
			break;
		}
	}

	// check SINICHIRO instance registered or not
	if (! chip)
	{
		/* it's a fresh one.
		 * now look for an empty slot and create a new card instance
		 * first, set the current configuration for this device
		 */
		{

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 5, 70)	/* FIXME: find right number */
			struct usb_host_config *config = dev->actconfig;
#else
			struct usb_config_descriptor *config = dev->actconfig;
#endif

			if (usb_reset_configuration(dev) < 0)
			{
				printk("SINICH-DRV : snd_SINICHIRO_probe : cannot reset configuration (value 0x%x)\n",
					   get_cfg_desc(config)->bConfigurationValue);
				goto __error;
			}
		}
		for (i = 0; i < SNDRV_CARDS; i++)
		{
			if (! usb_chip[i] &&
			    (vid[i] == -1 || vid[i] == dev->descriptor.idVendor) &&
			    (pid[i] == -1 || pid[i] == dev->descriptor.idProduct))
			{
				// create new SINICHIRO instance
				err = snd_SINICHIRO_create(dev, i, &chip);
				if (err < 0)
				{
					printk("SINICH-DRV : snd_SINICHIRO_probe : chip instance create error\n");
					goto __error;
				}
				break;
			}
		}
		if (! chip)
		{
			printk("SINICH-DRV : snd_SINICHIRO_probe : no available usb audio device\n");
			goto __error;
		}
	}

	chip->isopen = 0;

	usb_chip[chip->index] = chip;
	chip->num_interfaces++;
	up(&register_mutex);

	printk("SINICH_DRV : exit snd_SINICHIRO_probe\n");
	return chip;

 __error:

	printk("SINICH-DRV : snd_SINICHIRO_probe : __ERROR\n");

	up(&register_mutex);

	return NULL;
}


/******************************************************************************
 *	SINICHIRO disconnect function
 *
 *	parameter :
 *		dev			: usb device instance pointer
 *		ptr			: SINICHIRO chip instance pointer
 *
 *	return value :
 *		nothing
 */
static void snd_SINICHIRO_disconnect(struct usb_device *dev, void *ptr)
{
	snd_SINICHIRO_t *chip;

#if (SINICH_DRV_DEBUG > 0)
	printk("SINICH-DRV : snd_SINICHIRO_disconnect : \n");
#endif
	printk("SINICH_DRV : enter snd_SINICHIRO_disconnect\n");
	if (ptr == (void *)-1L)
	{
		return;
	}
	down(&register_mutex);
	chip = ptr;
	SINICHIRO_free_instance(chip);
	chip->shutdown = 1;
	chip->num_interfaces--;
	if (chip->num_interfaces <= 0)
	{
		usb_chip[chip->index] = NULL;
	}

	up(&register_mutex);
	printk("SINICH_DRV : exit snd_SINICHIRO_disconnect\n");
	
}


/******************************************************************************
 *	SINICHIRO probe function
 *
 *	parameter :
 *		intf		: usb interface structure pointer
 *		id			: usb ID number
 *
 *	return value :
 *		success(0) or error(<0)
 */
static int SINICHIRO_probe(struct usb_interface *intf,
						const struct usb_device_id *id)
{
	void *chip;
	int err;

#if (SINICH_DRV_DEBUG > 0)
	printk("SINICH-DRV : SINICHIRO_probe : \n");
#endif

	chip = snd_SINICHIRO_probe(interface_to_usbdev(intf), intf, id);


	if (chip)
	{
		dev_set_drvdata(&intf->dev, chip);
	}
	else
	{
		return -EIO;
	}

	//
	err = usb_register_dev(intf, &SINICHIRO_class);
	if (err)
	{
		usb_set_intfdata(intf, NULL);
		return -EIO;
	}

	return 0;
}


/******************************************************************************
 *	SINICHIRO disconnect function
 *
 *	parameter :
 *		intf		: usb interface structure pointer
 *
 *	return value :
 *		nothing
 */
static void SINICHIRO_disconnect(struct usb_interface *intf)
{
#if (SINICH_DRV_DEBUG > 0)
	printk("SINICH-DRV : SINICHIRO_disconnect : \n");
#endif

	snd_SINICHIRO_disconnect(interface_to_usbdev(intf),
						  dev_get_drvdata(&intf->dev));

	usb_deregister_dev(intf, &SINICHIRO_class);
}


/******************************************************************************
 *	SINICHIRO device ID
 */
static struct usb_device_id SINICHIRO_ids [] =
{
	// Multi-Channel USB : SINICHIRO
	{
		.match_flags	= USB_DEVICE_ID_MATCH_DEVICE,
		.idVendor		= 0x04EC,
		.idProduct		= 0x1121
	},
    { /* Terminating entry */ } 
};


MODULE_DEVICE_TABLE (usb, SINICHIRO_ids);


/******************************************************************************
 *	SINICHIRO usb driver structure
 */
static struct usb_driver SINICHIRO_driver =
{
//	.owner			= THIS_MODULE,
	.name			= "sinichdrv",
	.probe			= SINICHIRO_probe,
	.disconnect		= SINICHIRO_disconnect,
	.id_table		= SINICHIRO_ids,
};


/******************************************************************************
 *	SINICHIRO init function
 *
 *	parameter :
 *		nothing
 *
 *	return value :
 *		err	: success(0) or error(<0)
 */
static int __init SINICHIRO_init(void)
{
	int err = 0;

	// driver load message
	printk("SINICH-DRV : SINICHIRO_init : \n");

	// register USB device driver to system
	err = usb_register(&SINICHIRO_driver);
#if (SINICH_DRV_DEBUG > 0)
	printk("SINICH-DRV : SINICHIRO_init : register USB driver : err %d\n", err);
#endif

	return err;
}


/******************************************************************************
 *	SINICHIRO exit function
 *
 *	parameter :
 *		nothing
 *
 *	return value :
 *		nothing
 */
static void __exit SINICHIRO_cleanup(void)
{
	// driver unload message
	printk("SINICH-DRV : SINICHIRO_cleanup : \n");

	// de-register USB device driver from system
	usb_deregister(&SINICHIRO_driver);
}


module_init(SINICHIRO_init);
module_exit(SINICHIRO_cleanup);


/******************************************************************************
 *	for GPL
 */
MODULE_DESCRIPTION("TD-BD-USBIF sinich");
MODULE_LICENSE("GPL");
